@@ -1,31 +1,17 @@
-/*=============================================================================
-
-    This file is part of ARB.
-
-    ARB is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    ARB is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with ARB; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-=============================================================================*/
-/******************************************************************************
-
+/*
     Copyright (C) 2013 Fredrik Johansson
 
-******************************************************************************/
+    This file is part of Arb.
+
+    Arb is free software: you can redistribute it and/or modify it under
+    the terms of the GNU Lesser General Public License (LGPL) as published
+    by the Free Software Foundation; either version 2.1 of the License, or
+    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+*/
 
 #include "arb.h"
 
-#define ASSERT(cond) if (!(cond)) { flint_printf("FAIL: %d\n", __LINE__); abort(); }
+#define ASSERT(cond) if (!(cond)) { flint_printf("FAIL: %d\n", __LINE__); flint_abort(); }
 
 int main()
 {
@@ -607,6 +593,28 @@ int main()
         arb_get_abs_lbound_arf(b, indet_pos_rad, wp); ASSERT(arf_is_nan(b));
         arb_get_abs_lbound_arf(b, indet_inf_rad, wp); ASSERT(arf_is_nan(b));
 
+        arb_get_ubound_arf(b, zero, wp); ASSERT(arf_is_zero(b));
+        arb_get_ubound_arf(b, pos, wp); ASSERT(arf_sgn(b) > 0);
+        arb_get_ubound_arf(b, pos_inf, wp); ASSERT(arf_is_pos_inf(b));
+        arb_get_ubound_arf(b, neg_inf, wp); ASSERT(arf_is_neg_inf(b));
+        arb_get_ubound_arf(b, pos_inf_err, wp); ASSERT(arf_is_pos_inf(b));
+        arb_get_ubound_arf(b, neg_inf_err, wp); ASSERT(arf_is_neg_inf(b));
+        arb_get_ubound_arf(b, zero_pm_inf, wp); ASSERT(arf_is_pos_inf(b));
+        arb_get_ubound_arf(b, indet_exact, wp); ASSERT(arf_is_nan(b));
+        arb_get_ubound_arf(b, indet_pos_rad, wp); ASSERT(arf_is_nan(b));
+        arb_get_ubound_arf(b, indet_inf_rad, wp); ASSERT(arf_is_nan(b));
+
+        arb_get_lbound_arf(b, zero, wp); ASSERT(arf_is_zero(b));
+        arb_get_lbound_arf(b, neg, wp); ASSERT(arf_sgn(b) < 0);
+        arb_get_lbound_arf(b, pos_inf, wp); ASSERT(arf_is_pos_inf(b));
+        arb_get_lbound_arf(b, neg_inf, wp); ASSERT(arf_is_neg_inf(b));
+        arb_get_lbound_arf(b, pos_inf_err, wp); ASSERT(arf_is_pos_inf(b));
+        arb_get_lbound_arf(b, neg_inf_err, wp); ASSERT(arf_is_neg_inf(b));
+        arb_get_lbound_arf(b, zero_pm_inf, wp); ASSERT(arf_is_neg_inf(b));
+        arb_get_lbound_arf(b, indet_exact, wp); ASSERT(arf_is_nan(b));
+        arb_get_lbound_arf(b, indet_pos_rad, wp); ASSERT(arf_is_nan(b));
+        arb_get_lbound_arf(b, indet_inf_rad, wp); ASSERT(arf_is_nan(b));
+
         arf_clear(b);
     }
 
