@@ -30,21 +30,30 @@ void *
 _arb_mat_mul_thread(void * arg_ptr)
 {
     arb_mat_mul_arg_t arg = *((arb_mat_mul_arg_t *) arg_ptr);
-    slong i, j, k;
+    slong i, j, br, bc;
+    arb_ptr tmp;
+    TMP_INIT;
+
+    br = arg.br;
+    bc = arg.bc1 - arg.bc0;
+
+    TMP_START;
+    tmp = TMP_ALLOC(sizeof(arb_struct) * br * bc);
+
+    for (i = 0; i < br; i++)
+        for (j = 0; j < bc; j++)
+            tmp[j * br + i] = arg.B[i][arg.bc0 + j];
 
     for (i = arg.ar0; i < arg.ar1; i++)
     {
         for (j = arg.bc0; j < arg.bc1; j++)
         {
-            arb_mul(arg.C[i] + j, arg.A[i] + 0, arg.B[0] + j, arg.prec);
-
-            for (k = 1; k < arg.br; k++)
-            {
-                arb_addmul(arg.C[i] + j, arg.A[i] + k, arg.B[k] + j, arg.prec);
-            }
+            arb_dot(arg.C[i] + j, NULL, 0,
+                arg.A[i], 1, tmp + (j - arg.bc0) * br, 1, br, arg.prec);
         }
     }
 
+    TMP_END;
     flint_cleanup();
     return NULL;
 }
@@ -110,6 +119,7 @@ arb_mat_mul_threaded(arb_mat_t C, const arb_mat_t A, const arb_mat_t B, slong pr
 
         args[i].br = br;
         args[i].prec = prec;
+
         pthread_create(&threads[i], NULL, _arb_mat_mul_thread, &args[i]);
     }
 
@@ -121,4 +131,3 @@ arb_mat_mul_threaded(arb_mat_t C, const arb_mat_t A, const arb_mat_t B, slong pr
     flint_free(threads);
     flint_free(args);
 }
-
