@@ -294,28 +294,33 @@ void mag_add(mag_t z, const mag_t x, const mag_t y);
 void mag_add_lower(mag_t z, const mag_t x, const mag_t y);
 
 void mag_add_ui(mag_t z, const mag_t x, ulong y);
+void mag_add_ui_lower(mag_t res, const mag_t x, ulong y);
 
 void mag_add_ui_2exp_si(mag_t z, const mag_t x, ulong y, slong e);
 
 void mag_div(mag_t z, const mag_t x, const mag_t y);
+void mag_div_lower(mag_t z, const mag_t x, const mag_t y);
 
 MAG_INLINE void
-mag_mul_2exp_si(mag_t z, const mag_t x, slong y)
+mag_inv(mag_t res, const mag_t x)
 {
-    if (mag_is_special(x))
-    {
-        mag_set(z, x);
-    }
-    else
-    {
-        if (y >= ADD2_FAST_MIN && y <= ADD2_FAST_MAX)
-            _fmpz_add_fast(MAG_EXPREF(z), MAG_EXPREF(x), y);
-        else
-            fmpz_add_si(MAG_EXPREF(z), MAG_EXPREF(x), y);
-        MAG_MAN(z) = MAG_MAN(x);
-    }
+    mag_t t;
+    mag_init(t); /* no need to free */
+    mag_one(t);
+    mag_div(res, t, x);
 }
 
+MAG_INLINE void
+mag_inv_lower(mag_t res, const mag_t x)
+{
+    mag_t t;
+    mag_init(t); /* no need to free */
+    mag_one(t);
+    mag_div_lower(res, t, x);
+}
+
+void mag_mul_2exp_si(mag_t z, const mag_t x, slong y);
+
 MAG_INLINE void
 mag_mul_2exp_fmpz(mag_t z, const mag_t x, const fmpz_t y)
 {
@@ -468,7 +473,41 @@ mag_fast_add_2exp_si(mag_t z, const mag_t x, slong e)
     }
 }
 
+/* requires that x is positive and finite */
+#define MAG_SET_D_2EXP(man, exp, x, xexp) \
+    do { \
+        int __cexp; \
+        double __x; \
+        int __fix; \
+        mp_limb_t __man; \
+        __x = frexp((x), &__cexp); \
+        __man = (mp_limb_t)(__x * (double)(LIMB_ONE << MAG_BITS)) + 1; \
+        __fix = __man >> (MAG_BITS); \
+        __man = (__man >> __fix) + __fix; \
+        (man) = __man; \
+        (exp) = (xexp) + __cexp + __fix; \
+    } while (0);
+
+/* requires that x is positive and finite */
+#define MAG_SET_D_2EXP_LOWER(man, exp, x, xexp) \
+    do { \
+        int __cexp; \
+        double __x; \
+        int __fix; \
+        mp_limb_t __man; \
+        __x = frexp((x), &__cexp); \
+        __man = (mp_limb_t)(__x * (double)(LIMB_ONE << MAG_BITS)) - 1; \
+        __fix = __man < MAG_ONE_HALF; \
+        __man = (__man << __fix); \
+        (man) = __man; \
+        (exp) = (xexp) + __cexp - __fix; \
+    } while (0);
+
+void mag_set_d(mag_t z, double x);
+void mag_set_d_lower(mag_t z, double x);
 void mag_set_d_2exp_fmpz(mag_t z, double c, const fmpz_t exp);
+void mag_set_d_2exp_fmpz_lower(mag_t z, double c, const fmpz_t exp);
+
 void mag_set_fmpz_2exp_fmpz(mag_t z, const fmpz_t man, const fmpz_t exp);
 
 #include "fmpr.h"
@@ -499,6 +538,9 @@ mag_printd(const mag_t x, slong d)
 
 void mag_get_fmpq(fmpq_t y, const mag_t x);
 
+void mag_get_fmpz(fmpz_t res, const mag_t x);
+void mag_get_fmpz_lower(fmpz_t res, const mag_t x);
+
 int mag_cmp(const mag_t x, const mag_t y);
 
 int mag_cmp_2exp_si(const mag_t x, slong e);
@@ -542,14 +584,6 @@ _mag_vec_clear(mag_ptr v, slong n)
     flint_free(v);
 }
 
-MAG_INLINE void mag_set_d(mag_t z, double x)
-{
-    fmpz_t e;
-    fmpz_init(e);
-    mag_set_d_2exp_fmpz(z, x, e);
-    fmpz_clear(e);
-}
-
 double mag_get_d(const mag_t z);
 
 double mag_get_d_log2_approx(const mag_t x);
@@ -562,25 +596,40 @@ void mag_log1p(mag_t z, const mag_t x);
 
 void mag_log_ui(mag_t t, ulong n);
 
-void mag_exp_maglim(mag_t y, const mag_t x, slong maglim);
+void mag_log(mag_t z, const mag_t x);
+void mag_log_lower(mag_t z, const mag_t x);
+void mag_neg_log(mag_t z, const mag_t x);
+void mag_neg_log_lower(mag_t z, const mag_t x);
 
-MAG_INLINE void
-mag_exp(mag_t y, const mag_t x)
-{
-    mag_exp_maglim(y, x, 128);
-}
+void mag_exp(mag_t y, const mag_t x);
+void mag_exp_lower(mag_t y, const mag_t x);
+
+void mag_expinv(mag_t res, const mag_t x);
+void mag_expinv_lower(mag_t y, const mag_t x);
 
 void mag_expm1(mag_t y, const mag_t x);
 void mag_exp_tail(mag_t z, const mag_t x, ulong N);
 
-void mag_expinv(mag_t res, const mag_t x);
+void mag_sinh(mag_t y, const mag_t x);
+void mag_sinh_lower(mag_t y, const mag_t x);
+
+void mag_cosh(mag_t y, const mag_t x);
+void mag_cosh_lower(mag_t y, const mag_t x);
 
 void mag_pow_ui(mag_t z, const mag_t x, ulong e);
 void mag_pow_ui_lower(mag_t z, const mag_t x, ulong e);
 void mag_pow_fmpz(mag_t z, const mag_t x, const fmpz_t e);
+void mag_pow_fmpz_lower(mag_t z, const mag_t x, const fmpz_t e);
+
+void mag_const_pi(mag_t res);
+void mag_const_pi_lower(mag_t res);
+
+void mag_atan(mag_t res, const mag_t x);
+void mag_atan_lower(mag_t res, const mag_t x);
 
 void mag_fac_ui(mag_t z, ulong n);
 void mag_rfac_ui(mag_t z, ulong n);
+void mag_bin_uiui(mag_t res, ulong n, ulong k);
 
 /* TODO: test */
 void mag_bernoulli_div_fac_ui(mag_t z, ulong n);
@@ -589,7 +638,9 @@ void mag_bernoulli_div_fac_ui(mag_t z, ulong n);
 void mag_set_fmpz_2exp_fmpz_lower(mag_t z, const fmpz_t man, const fmpz_t exp);
 
 void mag_sqrt(mag_t y, const mag_t x);
+void mag_sqrt_lower(mag_t y, const mag_t x);
 void mag_rsqrt(mag_t y, const mag_t x);
+void mag_rsqrt_lower(mag_t y, const mag_t x);
 
 void mag_root(mag_t y, const mag_t x, ulong n);
 
@@ -607,12 +658,7 @@ void mag_set_ui(mag_t z, ulong x);
 void mag_set_ui_lower(mag_t z, ulong x);
 
 /* TODO: test functions below */
-MAG_INLINE void
-mag_set_ui_2exp_si(mag_t z, ulong v, slong e)
-{
-    mag_set_ui(z, v);
-    mag_mul_2exp_si(z, z, e);
-}
+void mag_set_ui_2exp_si(mag_t z, ulong v, slong e);
 
 MAG_INLINE void
 mag_set_fmpz(mag_t z, const fmpz_t x)
@@ -690,11 +736,6 @@ mag_div_fmpz(mag_t z, const mag_t x, const fmpz_t y)
     mag_clear(t);
 }
 
-MAG_INLINE void mag_const_pi(mag_t res)
-{
-    mag_set_ui_2exp_si(res, 843314857, -28);
-}
-
 MAG_INLINE slong mag_allocated_bytes(const mag_t x)
 {
     return fmpz_allocated_bytes(MAG_EXPREF(x));
